unit MainForm_;

{
  IMAP Export | by Proger_XP | in public domain
  http://proger.i-forge.net/ImapExport/ect

  This project uses Indy 10 components, not Indy 9 that Delphi 7 is packaged with.
  http://www.indyproject.org/sockets/Docs/Indy10Installation.EN.aspx
}

{$WARN UNIT_PLATFORM OFF}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ComCtrls, ExtCtrls, IdIOHandler,
  IdIOHandlerSocket, IdIOHandlerStack, IdSSL, IdSSLOpenSSL,
  IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient,
  IdExplicitTLSClientServerBase, IdMessageClient, IdIMAP4, Menus;

type
  TMainForm = class(TForm)
    edImapLogin: TEdit;
    GroupBox1: TGroupBox;
    edImapPassword: TEdit;
    cbImapUseSSL: TCheckBox;
    bbImapConnect: TBitBtn;
    edImapHost: TEdit;
    edImapPort: TEdit;
    GroupBox2: TGroupBox;
    cbImapFolders: TComboBox;
    lbImapFolderTotalMsg: TLabel;
    bbImapExport: TBitBtn;
    bbImapDisconnect: TSpeedButton;
    ProgressBar: TProgressBar;
    pnProgress: TPanel;
    lbProgressLegend: TLabel;
    Bevel1: TBevel;
    mmLog: TMemo;
    bbImapFolderShortcut: TSpeedButton;
    edImapExportDir: TEdit;
    Image1: TImage;
    pnImapExportDirExists: TPanel;
    Label1: TLabel;
    Imap: TIdIMAP4;
    SslSocket: TIdSSLIOHandlerSocketOpenSSL;
    IoSocket: TIdIOHandlerStack;
    bbImapFolderActions: TSpeedButton;
    pmImapFolderActions: TPopupMenu;
    miImapSelectExportDir: TMenuItem;
    Showmessagebyitsindex1: TMenuItem;
    pnImapSettings: TPanel;
    Setoutputfilename1: TMenuItem;
    cbSaveSettingsOnExit: TCheckBox;
    Skipmessagespriortodate1: TMenuItem;
    miSetDateOnExport: TMenuItem;
    procedure bbImapConnectClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure bbImapDisconnectClick(Sender: TObject);
    procedure cbImapFoldersSelect(Sender: TObject);
    procedure bbImapExportClick(Sender: TObject);
    procedure edImapExportDirKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure miImapSelectExportDirClick(Sender: TObject);
    procedure Showmessagebyitsindex1Click(Sender: TObject);
    procedure bbImapFolderActionsClick(Sender: TObject);
    procedure ImapWorkForPart(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCount: Int64);
    procedure ImapWorkBeginForPart(ASender: TObject; AWorkMode: TWorkMode;
      AWorkCountMax: Int64);
    procedure ImapWorkEndForPart(ASender: TObject; AWorkMode: TWorkMode);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormShow(Sender: TObject);
    procedure Setoutputfilename1Click(Sender: TObject);
    procedure Skipmessagespriortodate1Click(Sender: TObject);
    procedure cbSaveSettingsOnExitClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure ImapDisconnected(Sender: TObject);
  protected
    function LoadSettingsFrom(const FN: String): Boolean;
    procedure SaveSettingsTo(const FN: String);

    procedure UpdateUiState;
      procedure SetImapConnectState(SetEnabled: Boolean);
      procedure SetImapFoldersState(SetEnabled: Boolean);
    procedure UpdateWindowSize;

    procedure ClearLog;
      procedure LogInfo(Message: WideString; FormatArgs: array of const);
      procedure LogWarning(Message: WideString; FormatArgs: array of const);

    procedure BeginProgress(Legend: WideString; MaxProgress: Integer; InitialProgress: Integer = 0);
      procedure UpdateProgress(Current: Integer; Max: Integer = -1);  // -1 doesn't change max progress.
      procedure EndProgress;

    procedure ForceDisconnect;
    procedure ImapConnectTo(Host: String; Port: Integer; UseSSL: Boolean; Login, Password: String);
      procedure OnConnect;
        procedure FillImapFolders;
    procedure ShowMessageByIndex(Index: Integer);

    procedure SelectImapFolder(const Name: String); overload;
    procedure SelectImapFolder(Index: Word); overload;
    procedure SelectImapFolderShortcut(MenuItem: TMenuItem);
    procedure SetExportDir(Dir: WideString; IsSetByUser: Boolean);
    function GetMessagesForExport: TStrings;
      procedure ExportMessagesTo(ExportDir: WideString; Headers: TStrings);
  end;

var
  MainForm: TMainForm;

implementation

uses IdMailBox, IdMessage, IdEMailAddress, FileCtrl, ShellAPI, IniFiles,
  DateUtils;

const         
  EOLN = #13#10;
  DoNotAutoChangeImapExportDir = 1;
  MaxMessageFNLength = 100;
  MaxMessageSizeToExport = 50 * 1024 * 1024;  // 50 MiB.
  DefaultDateTime: TDateTime = 0;

type
  TImapWorkHandler = (NoWorkHandler, MessageRetrieval);
var
  ImapWorkHandler: TImapWorkHandler = (NoWorkHandler);
  SettingsFile: String = 'ImapExport.ini';
  SaveImapPassword: Boolean = False;
  DefaultImapFolder: String;
  MessageFN: String = '%INDEX%.eml';
  MessageStartDate: TDateTime;

{$R *.dfm}

function ToValidFN(const Str: String): String;
const
  WrongChars = '\/?*%:|"<>';
var
  I: Integer;
begin
  Result := Str;

  for I := 1 to Length(Result) do
    if IsDelimiter(WrongChars, Result, I) then
      Result[I] := '-';
end;

procedure TMainForm.bbImapConnectClick(Sender: TObject);
begin
  ImapConnectTo(edImapHost.Text, StrToInt(edImapPort.Text), cbImapUseSSL.Checked,
                edImapLogin.Text, edImapPassword.Text);
end;

procedure TMainForm.ImapConnectTo(Host: String; Port: Integer; UseSSL: Boolean; Login, Password: String);
begin
  ForceDisconnect;

  Imap.Host := Host;
    Imap.Port := Port;
  Imap.Username := Login;
    Imap.Password := Password;

    if UseSSL then
    begin
      Imap.IOHandler := SslSocket;
      Imap.UseTLS := utUseImplicitTLS;
    end
      else
      begin
        Imap.IOHandler := IoSocket;
        Imap.UseTLS := utNoTLSSupport;
      end;

  try
    Screen.Cursor := crHourGlass;
    Imap.Connect;
  finally
    Screen.Cursor := crDefault;
  end;

  OnConnect;                
  SelectImapFolder(DefaultImapFolder);
  cbImapFolders.SetFocus;
end;

  procedure TMainForm.OnConnect;
  begin
    FillImapFolders;
    UpdateUiState;
  end;

    procedure TMainForm.FillImapFolders;
    begin
      Imap.ListMailBoxes( TStringList(cbImapFolders.Items) );
      SelectImapFolder(0);
    end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  SettingsFile := ExtractFilePath(ParamStr(0)) + '\' + SettingsFile;
  ClearLog;
  EndProgress;
  UpdateUiState;
  SetExportDir('', False);
  SelectImapFolderShortcut(miImapSelectExportDir);
end;

function TMainForm.LoadSettingsFrom(const FN: String): Boolean;
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(FN);
  try
    edImapHost.Text := Ini.ReadString('Connection', 'Host', edImapHost.Text);
    edImapPort.Text := Ini.ReadString('Connection', 'Port', edImapPort.Text);
    edImapLogin.Text := Ini.ReadString('Connection', 'Login', edImapLogin.Text);
    edImapPassword.Text := Ini.ReadString('Connection', 'Password', edImapPassword.Text);
    cbImapUseSSL.Checked := Ini.ReadBool('Connection', 'UseSSL', cbImapUseSSL.Checked);

    DefaultImapFolder := Ini.ReadString('General', 'DefaultFolder', DefaultImapFolder);
    MessageFN := Ini.ReadString('General', 'MessageFileName', MessageFN);
    MessageStartDate := Ini.ReadDateTime('General', 'SkipPriorToDate', MessageStartDate);
    miSetDateOnExport.Checked := Ini.ReadBool('General', 'SetSkipDateOnExport', miSetDateOnExport.Checked);
    Result := Ini.ReadBool('General', 'ConnectOnRun', False);
  finally
    Ini.Free;
  end;
end;

procedure TMainForm.SaveSettingsTo(const FN: String);
var
  Ini: TIniFile;
begin
  Ini := TIniFile.Create(FN);
  try
    Ini.WriteString('Connection', 'Host', edImapHost.Text);
    Ini.WriteString('Connection', 'Port', edImapPort.Text);
    Ini.WriteString('Connection', 'Login', edImapLogin.Text);
    if SaveImapPassword then
      Ini.WriteString('Connection', 'Password', edImapPassword.Text);
    Ini.WriteBool('Connection', 'UseSSL', cbImapUseSSL.Checked);

    if (DefaultImapFolder <> '') and Imap.Connected then
      DefaultImapFolder := Imap.MailBox.Name;

    Ini.WriteString('General', 'DefaultFolder', DefaultImapFolder);
    Ini.WriteString('General', 'MessageFileName', MessageFN);
    Ini.WriteDateTime('General', 'SkipPriorToDate', MessageStartDate);
    Ini.WriteBool('General', 'SetSkipDateOnExport', miSetDateOnExport.Checked);
    Ini.WriteBool('General', 'ConnectOnRun', Ini.ReadBool('General', 'ConnectOnRun', False));
  finally
    Ini.Free;
  end;
end;

procedure TMainForm.UpdateUiState;
begin
  SetImapConnectState(not Imap.Connected);
  SetImapFoldersState(Imap.Connected);
end;

  procedure TMainForm.SetImapConnectState(SetEnabled: Boolean);
  begin
    edimaplogin.Enabled := SetEnabled;
    edimapPassword.Enabled := SetEnabled;
    edimaphost.Enabled := SetEnabled;
    edimapport.Enabled := SetEnabled;
    cbimapusessl.Enabled := SetEnabled;

    bbImapConnect.Enabled := SetEnabled;
      bbImapConnect.Default := SetEnabled;
    bbImapDisconnect.Enabled := not SetEnabled;
  end;

  procedure TMainForm.SetImapFoldersState(SetEnabled: Boolean);
  begin
    cbImapFolders.Enabled := SetEnabled;

    bbImapExport.Enabled := SetEnabled;
      bbImapExport.Default := SetEnabled;
    bbImapFolderShortcut.Enabled := SetEnabled;
    edImapExportDir.Enabled := SetEnabled;
    bbImapFolderActions.Enabled := SetEnabled;

    if not SetEnabled and not Imap.Connected then
    begin
      cbImapFolders.Items.Clear;
      cbImapFolders.Items.Add('First connect to a host...');
      cbImapFolders.ItemIndex := 0;

      lbImapFolderTotalMsg.Caption := 'Enter connection info and press Connect...';
    end;
  end;

procedure TMainForm.bbImapDisconnectClick(Sender: TObject);
begin
  ForceDisconnect;
end;
               
procedure TMainForm.ForceDisconnect;
begin
  if Imap.Connected then
    try
      Imap.Disconnect(True);
    except
      Imap.Disconnect(False);
    end;
end;

procedure TMainForm.cbImapFoldersSelect(Sender: TObject);
  function ShortExportDir(Folder: String): String;
  const
    Letters = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
  begin
    Result := Folder;

    if (Length(Folder) > 5) and (StrLIComp(PChar(Folder), 'inbox', 5) = 0) and not IsDelimiter(Letters, Folder, 6) then
      Result := Copy(Result, 7, MaxInt);
      
    Result := ToValidFN(Result);
  end;

begin
  try
    try
      Screen.Cursor := crHourGlass;

      if not Imap.SelectMailBox( TComboBox(Sender).Text ) then
        raise Exception.Create('');
    finally
      Screen.Cursor := crDefault;
    end;
  except
    MessageBox(Handle, 'Cannot select this folder due to a server error.', 'IMAP Export', mb_IconStop);

    if TComboBox(Sender).ItemIndex = 0 then
      ForceDisconnect
      else
        SelectImapFolder(0);

    Exit;
  end;

  try
    lbImapFolderTotalMsg.Caption := Format('%d messages in %1.1f MiB.',
                                           [Imap.MailBox.TotalMsgs, Imap.RetrieveMailBoxSize / 1024 / 1024]);
  except
    lbImapFolderTotalMsg.Caption := '';
  end;

  bbImapExport.Enabled := Imap.MailBox.TotalMsgs <> 0;
  SetExportDir( ExtractFilePath(ParamStr(0)) + ShortExportDir(Imap.MailBox.Name), False );
end;

  procedure TMainForm.SelectImapFolder(const Name: String);
  var
    Index: Integer;
  begin
    Index := cbImapFolders.Items.IndexOf(Name);
    if Index <> -1 then
      SelectImapFolder(Index);
  end;

  procedure TMainForm.SelectImapFolder(Index: Word);
  begin
    cbImapFolders.ItemIndex := Index;
    cbImapFolders.OnSelect(cbImapFolders);
  end;

procedure TMainForm.bbImapExportClick(Sender: TObject);
begin
  Randomize;    // for %RANDOM%
  SetImapConnectState(False);
  SetImapFoldersState(False);

  try
    ExportMessagesTo( edImapExportDir.Text, GetMessagesForExport );
  finally
    UpdateUiState;
  end;
end;

  function TMainForm.GetMessagesForExport: TStrings;
    function Matches(Message: TIdMessage): Boolean;
    begin
      Result := CompareDateTime(Message.Date, MessageStartDate) > 0;
    end;

  var
    I, MessageSize: Integer;
    Message: TIdMessage;
  begin
    BeginProgress('Retreiving message list...', Imap.MailBox.TotalMsgs);

    if not SameDateTime(MessageStartDate, DefaultDateTime) then
      LogInfo('Ignoring all messages prior to %s.', [FormatDateTime('c', MessageStartDate)]);

    try
      Result := TStringList.Create;
      Result.BeginUpdate;

      try
        for I := 1 to Imap.MailBox.TotalMsgs do
        begin
          Message := TIdMessage.Create(NIL);

          try
            // todo: in future here we can put a filter to export specific messages rather than all existing.
            MessageSize := Imap.RetrieveMsgSize(I);

            if MessageSize <= MaxMessageSizeToExport then
              if Imap.RetrieveHeader(I, Message) then
              begin
                if Matches(Message) then
                  Result.AddObject( Message.Headers.Values['Subject'], TObject(I) );
              end
                else
                  LogWarning('Could not retrieve header of message #%d.', [I])
              else
                LogWarning('Size of message #%d (%1.1fK) is larger than export max (%1.0fK).',
                           [I, MessageSize / 1024, MaxMessageSizeToExport / 1024]);

            UpdateProgress(I);
          finally
            Message.Free;
          end;
        end;
      except
        Result.Free;
        raise;
      end;
    finally
      EndProgress;
    end;
  end;

    procedure SaveMessageTo(const FileName: WideString; Message: TIdMessage); forward;
    function MessageFileNameFor(Message: TIdMessage; Index: Integer): String; forward;
    function MessageFileNameIn(Dir, MessageFN: String): String; forward;

  procedure TMainForm.ExportMessagesTo(ExportDir: WideString; Headers: TStrings);
  var
    I, ExportedCount, MessageIndex: Integer;
    Message: TIdMessage;
    RetrievedOK: Boolean;
  begin
    if Headers.Count <> 0 then
      if ForceDirectories(ExportDir) then
      begin
        BeginProgress('Exporting messages...', Headers.Count);

        try
          ExportedCount := 0;
          ExportDir := IncludeTrailingPathDelimiter(ExportDir);

          for I := 0 to Headers.Count - 1 do
          begin
            Message := TIdMessage.Create(NIL);
            try
              MessageIndex := Integer(Headers.Objects[I]);

              try
                RetrievedOK := Imap.Retrieve(MessageIndex, Message);
              except
                RetrievedOK := False;
              end;

              if RetrievedOK then
              begin
                SaveMessageTo(MessageFileNameIn(ExportDir, MessageFileNameFor(Message, MessageIndex)), Message);
                Inc(ExportedCount);
              end
                else
                  LogWarning('Could not retrieve message #%d.', [I]);

              UpdateProgress(I + 1);
            finally
              Message.Free;
            end;
          end;

          LogInfo('%d messages have been exported to %s.',
                  [ExportedCount, ExcludeTrailingPathDelimiter(ExportDir)]);
        finally
          EndProgress;
        end;
      end
        else
          raise Exception.CreateFmt('Cannot create export directory %s.', [ExportDir]);
  end;

    procedure SaveMessageTo(const FileName: WideString; Message: TIdMessage);
    var
      OutFile: TFileStream;
    begin
      ForceDirectories(ExtractFilePath(FileName));

      OutFile := TFileStream.Create(FileName, fmCreate);
      try
        Message.SaveToStream(OutFile);
      finally
        OutFile.Free;
      end;
    end;

    function MessageFileNameFor(Message: TIdMessage; Index: Integer): String;
      procedure Replace(const VarName: String; S: String);
      begin
        if S = '' then
          S := '-';
        Result := StringReplace(Result, '%' + VarName + '%', S, [rfReplaceAll]);
      end;

      procedure ReplaceAddrList(const VarName: String; List: TIdEmailAddressList);
      begin
        if List.Count > 0 then
          Replace(VarName, List[0].Address)
          else
            Replace(VarName, '-');
      end;

    var
      Flags: String;
    begin
      Result := MessageFN;

      Replace('INDEX', IntToStr(Index));
      Replace('MSGID', Message.MsgId);
      Replace('RANDOM', IntToStr(Random(9998) + 1));
      Replace('UID', Message.UID);
      Replace('CHARSET', Message.CharSet);
      Replace('FROM', Message.From.Address);
      Replace('SUBJECT', Message.Subject);
      Replace('PRIORITY', IntToStr(Ord(Message.Priority)));

      ReplaceAddrList('TO', Message.Recipients);
      ReplaceAddrList('CC', Message.CCList);
      ReplaceAddrList('BCC', Message.BccList);

      Replace('LDATE', FormatDateTime(LongDateFormat, Message.Date));
      Replace('SDATE', FormatDateTime(ShortDateFormat, Message.Date));
      Replace('LTIME', FormatDateTime(LongTimeFormat, Message.Date));
      Replace('STIME', FormatDateTime(ShortTimeFormat, Message.Date));

      Flags := '';

        if mfAnswered in Message.Flags then   Flags := Flags + 'A';
        if mfFlagged in Message.Flags then    Flags := Flags + 'F';
        if mfDeleted in Message.Flags then    Flags := Flags + 'R';
        if mfDraft in Message.Flags then      Flags := Flags + 'D';
        if mfSeen in Message.Flags then       Flags := Flags + 'S';
        if mfRecent in Message.Flags then     Flags := Flags + 'N';

      Replace('FLAGS', Flags);

      if Length(Result) > MaxMessageFNLength then
        Result := Copy(Result, 0, MaxMessageFNLength) + ExtractFileExt(MessageFN);

      Result := ToValidFN(Result);
    end;

    function MessageFileNameIn(Dir, MessageFN: String): String;
    var
      I: Integer;
    begin
      Dir := IncludeTrailingPathDelimiter(Dir);
      Result := Dir + MessageFN;
      I := 2;

      if FileExists(Result) then
        repeat
          Result := Dir + ChangeFileExt(MessageFN, '') + ' (' + IntToStr(I) + ')' + ExtractFileExt(MessageFN);
          Inc(I);
        until not FileExists(Result);
    end;

procedure TMainForm.BeginProgress(Legend: WideString; MaxProgress: Integer; InitialProgress: Integer = 0);
begin
  Screen.Cursor := crAppStart;

  lbProgressLegend.Caption := Legend;
  ProgressBar.Position := InitialProgress;
  ProgressBar.Max := MaxProgress;

  pnProgress.Visible := True;

  UpdateWindowSize;
end;

  procedure TMainForm.UpdateProgress(Current, Max: Integer);
  begin
    if Max >= 0 then
      ProgressBar.Max := Max;
    ProgressBar.Position := Current;

    Application.ProcessMessages;
  end;

  procedure TMainForm.EndProgress;
  begin
    Screen.Cursor := crDefault;
    pnProgress.Visible := False;
    UpdateWindowSize;
  end;

procedure TMainForm.ClearLog;
begin
  mmLog.Clear;
  mmLog.Visible := False;

  UpdateWindowSize;
end;

  procedure TMainForm.LogInfo(Message: WideString; FormatArgs: array of const);
  begin
    mmLog.Lines.Add( 'INFO: ' + Format(Message, FormatArgs) );
    mmLog.Visible := mmLog.Lines.Count > 0;

    UpdateWindowSize;
  end;

  procedure TMainForm.LogWarning(Message: WideString; FormatArgs: array of const);
  begin
    mmLog.Lines.Add( 'WARN: ' + Format(Message, FormatArgs) );
    mmLog.Visible := mmLog.Lines.Count > 0;

    UpdateWindowSize;
  end;

procedure TMainForm.miImapSelectExportDirClick(Sender: TObject);
var
  Caption, Dir: String;
begin
  SelectImapFolderShortcut( TMenuItem(Sender) );

  Caption := Format('Select where to put at most %d messages exported from %s',
                    [Imap.MailBox.TotalMsgs, Imap.MailBox.Name]);

  if SelectDirectory(Caption, '', Dir) then
    SetExportDir(Dir, True);
end;

  procedure TMainForm.SetExportDir(Dir: WideString; IsSetByUser: Boolean);
  begin
    if IsSetByUser or (edImapExportDir.Tag <> DoNotAutoChangeImapExportDir) then
    begin
      edImapExportDir.Text := Dir;
      edImapExportDir.SelStart := Length(Dir);
      pnImapExportDirExists.Visible := DirectoryExists(Dir);
    end;

    if IsSetByUser then
      edImapExportDir.Tag := DoNotAutoChangeImapExportDir;
  end;

procedure TMainForm.edImapExportDirKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if Key >= Ord(' ') then
    SetExportDir( TEdit(Sender).Text, True );
end;

procedure TMainForm.Showmessagebyitsindex1Click(Sender: TObject);
var
  Index: String;
begin
  SelectImapFolderShortcut( TMenuItem(Sender) );

  Index := '1';

  if InputQuery('IMAP Export', 'Enter message index to show (first message is #1):', Index) then
    ShowMessageByIndex( StrToInt(Index) );
end;

  procedure TMainForm.ShowMessageByIndex(Index: Integer);
  var
    Message: TIdMessage;
    MessagesDir, MessageFile: WideString;
  begin
    if Index < 1 then
      raise Exception.CreateFmt('Message index must be 1 or greater but %d was passed.', [Index]);

    Message := TIdMessage.Create(NIL);
    try
      try
        ImapWorkHandler := MessageRetrieval;
        if not Imap.Retrieve(Index, Message) then
          raise Exception.Create('');
      except
        raise Exception.CreateFmt('Could not retirieve message with index #%d due to a server error -' +
                                  ' maybe it doesn''t exist?', [Index]);
      end;

      MessagesDir := IncludeTrailingPathDelimiter( ExpandFileName('by index') );

      if not DirectoryExists(MessagesDir) then
        try
          ForceDirectories(MessagesDir);
        except
          raise Exception.CreateFmt('Could not create directory for messages-by-index: %s.', [MessagesDir]);
        end;

      MessageFile := MessagesDir + IntToStr(Index) + '.eml';

      Message.SaveToFile(MessageFile);

      if ShellExecuteW(Handle, '', PWideChar(MessageFile), '', '', SW_NORMAL) <= 32 then
        raise Exception.CreateFmt('Message is saved to %s'#10'but there is no associated' +
                                  ' program with .EML files to run.', [MessageFile]);
    finally
      Message.Free;
      ImapWorkHandler := NoWorkHandler;
    end;
  end;

procedure TMainForm.bbImapFolderActionsClick(Sender: TObject);
begin
  miSetDateOnExport.Enabled := cbSaveSettingsOnExit.Checked;

  with TWinControl(Sender).BoundsRect do
    with TWinControl(Sender).Parent.ClientToScreen( Point(Left, Bottom) ) do
      pmImapFolderActions.Popup(X, Y);
end;

  procedure GrayscaleBitmap(Bitmap: TBitmap); forward;

procedure TMainForm.SelectImapFolderShortcut(MenuItem: TMenuItem);
  procedure AssignAndAddGrayedGlyphFrom(Source, Dest: TBitmap);
  begin
    { This is a simplified generator. Two of its shortcomings (not critical for this
      project at least) are that when bitmap's transparent color (lower-left corner) is
      converted to gray it flood fills it - but without checks if there are any colors
      that at first were not like that transparent one but after grayscaling they both
      got the same RGB; nor does it fills other possible transparent pixels, only filling
      lower-left,top-left and top-right corners. }

    Dest.Assign(Source);
      Dest.Width := 32;
      Dest.Canvas.CopyRect( Rect(16, 0, 32, 16), Dest.Canvas, Rect(0, 0, 16, 16) );

    GrayscaleBitmap(Dest);

    Dest.Canvas.CopyRect( Rect(0, 0, 16, 16), Source.Canvas, Rect(0, 0, 16, 16) );

    Dest.Canvas.Brush.Style := bsSolid;
    Dest.Canvas.Brush.Color := Source.Canvas.Pixels[0, 15];
      Dest.Canvas.FloodFill(16, 15, Dest.Canvas.Pixels[16, 15], fsSurface );
      Dest.Canvas.FloodFill(16, 0, Dest.Canvas.Pixels[16, 0], fsSurface );
      Dest.Canvas.FloodFill(31, 0, Dest.Canvas.Pixels[31, 0], fsSurface );
  end;

begin
  if MenuItem.InheritsFrom(TMenuItem) then
  begin
    MenuItem.Default := True;
    bbImapFolderShortcut.OnClick := MenuItem.OnClick;
    AssignAndAddGrayedGlyphFrom( MenuItem.Bitmap, bbImapFolderShortcut.Glyph );
  end;
end;

  // got from http://exampledelphi.com/delphi.php/graphic/convert-bitmap-image-to-grayscale/
  procedure GrayscaleBitmap(Bitmap: TBitmap);
  var
    I, J: Integer;
    Grayshade, Red, Green, Blue: Byte;
    PixelColor: Longint;
  begin
    with Bitmap do
      for I := 0 to Width - 1 do
        for J := 0 to Height - 1 do
        begin
          PixelColor := ColorToRGB(Canvas.Pixels[I, J]);
            Red        := PixelColor;
            Green      := PixelColor shr 8;
            Blue       := PixelColor shr 16;
            Grayshade  := Round(0.3 * Red + 0.6 * Green + 0.1 * Blue);
          Canvas.Pixels[I, J] := RGB(Grayshade, Grayshade, Grayshade);
        end;
  end;

procedure TMainForm.ImapWorkForPart(ASender: TObject; AWorkMode: TWorkMode;
  AWorkCount: Int64);
begin
  if ImapWorkHandler <> NoWorkHandler then
    UpdateProgress(AWorkCount);
end;

procedure TMainForm.ImapWorkBeginForPart(ASender: TObject;
  AWorkMode: TWorkMode; AWorkCountMax: Int64);
begin
  if ImapWorkHandler = MessageRetrieval then
    BeginProgress('Downloading message...', AWorkCountMax);
end;

procedure TMainForm.ImapWorkEndForPart(ASender: TObject;
  AWorkMode: TWorkMode);
begin
  if ImapWorkHandler <> NoWorkHandler then
    EndProgress;
end;

procedure TMainForm.UpdateWindowSize;
begin
  pnImapSettings.AutoSize := True;
  pnImapSettings.AutoSize := False;

  // we don't need additional space when progress panel is on because of its BorderWidth.
  if mmLog.Visible and not pnProgress.Visible then
    pnImapSettings.Height := pnImapSettings.Height + 8;
end;

procedure TMainForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if mmLog.Visible and miSetDateOnExport.Checked then
    MessageStartDate := Now;

  if cbSaveSettingsOnExit.Checked then
    SaveSettingsTo(SettingsFile)
    else
      DeleteFile(SettingsFile);

  ForceDisconnect;
end;

procedure TMainForm.FormShow(Sender: TObject);
var
  AutoConnect: Boolean;
begin
  if FileExists(SettingsFile) then
  begin
    cbSaveSettingsOnExit.Checked := True;
    AutoConnect := LoadSettingsFrom(SettingsFile);
    SaveImapPassword := edImapPassword.Text <> '';

    UpdateUiState;

    if AutoConnect then
      bbImapConnect.Click
      else if edImapPassword.Text = '' then
        edImapPassword.SetFocus;

    SelectImapFolder(DefaultImapFolder);
  end;
end;

procedure TMainForm.Setoutputfilename1Click(Sender: TObject);
const
  Prompt =
    'If name contains ''\''s directories are created as necessary.' + EOLN +
    'Symbols not valid in file names are replaced with ''-''.' + EOLN +
    'If resulting file name becomes longer than %MAXFNLEN% characters it is truncated.' + EOLN +
    'Substitution variables' + EOLN +
    EOLN +
    '%INDEX%'#9'- message index (1-based)' + EOLN +
    '%RANDOM%'#9'- random number (1-9999)' + EOLN +
    '%MSGID%'#9'- message ID' + EOLN +
    '%UID%'#9#9'- UID, usually empty (''-'')' + EOLN +
    '%CHARSET%'#9'- body character set or ''-''' + EOLN +
    '%FROM%'#9'- sender''s address' + EOLN +
    '%SUBJECT%'#9'- message subject line' + EOLN +
    '%PRIORITY%'#9'- message priority (0 for highest .. 4 for lowest)' + EOLN +
    '%TO%'#9#9'- message first recipient or ''-''' + EOLN +
    '%CC%'#9#9'- message first CC recipient or ''-''' + EOLN +
    '%BCC%'#9#9'- message first BCC (hidden CC) recipient or ''-''' + EOLN +
    '%LDATE%'#9'- date of message in long format' + EOLN +
    '%SDATE%'#9'- date of message in short format' + EOLN +
    '%LTIME%'#9'- time of message in long format' + EOLN +
    '%STIME%'#9'- time of message in short format' + EOLN +
    '%FLAGS%'#9'- message flags: A (answered), F (flagged), R (deleted), D (draft),' +
                               ' S (seen), N (recent), ''-'' (none of these)' + EOLN +
    EOLN +
    'Enter empty string to restore default %INDEX%.eml.';
var
  S: String;
begin
  SelectImapFolderShortcut( TMenuItem(Sender) );
  S := MessageFN;

  if InputQuery('New file name pattern for exported messages', StringReplace(Prompt, '%MAXFNLEN%', IntToStr(MaxMessageFNLength), []), S) then
    if S = '' then
      MessageFN := '%INDEX%.eml'
      else
        MessageFN := S;
end;

procedure TMainForm.Skipmessagespriortodate1Click(Sender: TObject);
const
  Prompt =
    'Format depends on your system''s locale settings.' + EOLN +
    'You can enter date with possible time. Date is 2 or 3 digits (%s).' + EOLN +
    'Time is 2 or 3 digits with optional AM/PM (%s).' + EOLN +
    EOLN +
    'Enter an empty string to reset the filter (export all messages).' + EOLN +
    'Enter "now" to set to current date and time.' + EOLN +
    EOLN +
    'If this is specified when exporting starts a message will appear with the recognized date and time.';
var
  FmtPrompt, S: String;
begin
  SelectImapFolderShortcut( TMenuItem(Sender) );
  FmtPrompt := Format(Prompt, [FormatDateTime(ShortDateFormat, Now), FormatDateTime(LongTimeFormat, Now)]);

  if not SameDateTime(MessageStartDate, DefaultDateTime) then
    S := FormatDateTime('c', MessageStartDate);

  if InputQuery('Skip messages prior to date when exporting', FmtPrompt, S) then
    try
      if S = '' then
        MessageStartDate := DefaultDateTime
        else if LowerCase(S) = 'now' then
          MessageStartDate := Now
          else
            MessageStartDate := StrToDateTime(S);
    except
      on EConvertError do
        MessageBox(Handle, 'Cannot recognize your date/time string.', 'Invalid format', mb_IconStop);
    end;
end;

procedure TMainForm.cbSaveSettingsOnExitClick(Sender: TObject);
begin
  if not FileExists(SettingsFile) then
    SaveImapPassword := True;
end;

procedure TMainForm.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
const
  Msg = 'You are currently connected to the server. Terminate and leave?';
begin
  if (Key = VK_ESCAPE) and (not Imap.Connected or (MessageBox(Handle, Msg, 'Still connected', mb_IconWarning or mb_YesNo) = idYes)) then
    Close;
end;

procedure TMainForm.ImapDisconnected(Sender: TObject);
begin
  UpdateUiState;
end;

end.
